import type { Logger } from "./logger.js";
export type HighlightManagerOptions<Anchor, Data> = {
    /** The container to which the manager will add highlight elements. */
    container: HTMLElement;
    /** A function transforming an anchor into one or more Ranges. */
    getRangesFromAnchor: GetRangesFromAnchorFunction<Anchor>;
    /** A function returning one or more extra class names to apply to a highlight's
     * elements. Called initially to get a highlight's classNames. After creating a highlight, call
     * updateHighlightsClassNames to refresh class names based on this callback.
     *
     * If omitted, a class name is added based on the highlight's index and the default number of
     * hightlight colors.
     */
    getHighlightClassNames?: GetHighlightClassNamesFunction<Data>;
    /** The class to apply to all highlight elements. If omitted a default value is used. */
    highlightClass?: string;
    /** The class to apply to hovered highlights. If omitted a default value is used. */
    hoverClass?: string;
    /** The class to apply to focused highlights. If omitted a default value is used. */
    focusClass?: string;
    /** The logger to use. If omitted, window.console is used. */
    logger?: Logger;
};
type GetRangesFromAnchorFunction<Anchor> = (anchor: Anchor) => Range[];
type GetHighlightClassNamesFunction<Data> = (highlightData: Data, index: number) => string[];
/** A class that manages a collection of highlights on a page. */
declare class HighlightManager<Anchor, Data> {
    private readonly options;
    /** A unique ID for this highlight manager. Used to determine if a highlight belongs to this manager. */
    private readonly highlightManagerId;
    private readonly highlights;
    private sortedHighlightElements;
    constructor(options: HighlightManagerOptions<Anchor, Data>);
    /** Notifies the manager that highlights matching the selector must update their classNames. */
    updateHighlightsClassNames(selector: HighlightSelector<Data>): string[][];
    private updateHighlightClassNames;
    /**
     * Scrolls to the first element of the first highlight matching the selector and
     * applies the focus style to it.
     */
    focusHighlight(selector: HighlightSelector<Data>): void;
    private applyFocusStyle;
    private applyHoverStyle;
    /**
     * Creates a new highlight. If the anchor produces ranges that are
     * co-extensive with an existing highlight, information for that highlight
     * is returned. Note that this returned object is not guaranteed to be
     * identical to the one that was returned previously, but they will have the
     * same `id`.
     */
    createHighlight(anchor: Anchor, data: Data, handlers?: HighlightHandlers<Data>): Highlight<Anchor, Data>;
    private getCoextensiveHighlight;
    private updateStyles;
    private addEventListeners;
    private insertSortedElement;
    private removeSortedElement;
    private handleMouseMove;
    private handleClick;
    private document;
    private window;
    private getHighestHighlightElementAtPoint;
    private handleResize;
    private updateAllHighlightElements;
    /**
     * Updates the highlight's elements to have the right coordinates. Adds
     * new elements as necessary, and hides unnecessary ones.
     */
    private updateHighlightElements;
    private createElementForHighlight;
    private combineAdjacentRects;
    /** Remove the highlight matching `highlight.id`. */
    removeHighlight(highlight: Highlight<Anchor, Data>): void;
    /** Remove highlights matching the selector. */
    removeHighlights(selector: HighlightSelector<Data>): void;
    /** Remove all highlights. */
    removeAllHighlights(): void;
    /** For testing: returns an array of the highlight's elements' bounding client rects. */
    getElementBoundingClientRects({ id }: Highlight<Anchor, Data>): DOMRect[];
    /** For testing: returns the count of the highlight's elements. */
    getElementCount({ id }: Highlight<Anchor, Data>): number;
}
export interface HighlightHandlers<Data> {
    onClick: (highlightData: Data) => void;
}
export { HighlightManager };
/** A datatype exposing readonly Highlight information. */
export interface Highlight<Anchor, Data> {
    /** A unique identifier for the highlight. */
    readonly id: string;
    /** The object that anchors the highlight to the document */
    readonly anchor: Readonly<Anchor>;
    /** Arbitrary data to associate with the highlight. Can be used to identify the highlight via callbacks. */
    readonly data: Readonly<Data>;
    /** The class names applied to the highlight. */
    readonly classNames: ReadonlyArray<string>;
}
export declare const classNameIndexPlaceholder = "{index}";
type HighlightSelector<Data> = (highlightData: Data) => boolean;
